"""
Multiply
========

"""

import numpy as np

from .fitness_normalizer import FitnessNormalizer


class Multiply(FitnessNormalizer):
    """
    Multiplies the fitness values by some coefficient.

    Examples
    --------
    *Multiplying Fitness Values by a Set of Coefficients*

    In this example, assume that each fitness value consists of a
    :class:`tuple` of numbers, each representing a different property
    of the molecule, and each contributing to the final fitness value.
    The properties can be anything, such as  energy, number of atoms
    or diameter.

    If your final fitness value depends on the combination of these
    properties, you will probably first want to scale them with
    :class:`.DivideByMean`. Once this is done, you may want to
    multiply each property by some coefficient, which reflects its
    relative importance to the final fitness value. For example
    if you multiply the value of one property by ``1`` and another
    by ``2``, the second will contribute twice as much to the
    final fitness value, assuming that you get the final fitness value
    by using the :class:`.Sum` normalizer after :class:`.Multiply`.

    Giving a concrete example

    .. testcode:: multiplying-fitness-values-by-a-set-of-coefficients

        import stk
        import numpy as np

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(1, 1, 1),
                normalized=False,
            ),
        )

        normalizer = stk.Multiply((1, 2, 3))
        normalized_population = tuple(normalizer.normalize(population))
        normalized_record, = normalized_population
        assert np.all(np.equal(
            normalized_record.get_fitness_value(),
            (1, 2, 3),
        ))

    *Selectively Normalizing Fitness Values*

    Sometimes, you only want to normalize some members of a population,
    for example if some do not have an assigned fitness value,
    because the fitness calculation failed for whatever reason.
    You can use the `filter` parameter to exclude records from the
    normalization

    .. testcode:: selectively-normalizing-fitness-values

        import stk
        import numpy as np

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(1, 1, 1),
                normalized=False,
            ),
            # This will have a fitness value of None.
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ),
        )

        normalizer = stk.Multiply(
            coefficient=(1, 2, 3),
            # Only normalize values which are not None.
            filter=lambda population, record:
                record.get_fitness_value() is not None,
        )

        # Calling normalizer.normalize() will return a new
        # population holding the molecule records with normalized
        # fitness values.
        normalized_population = tuple(normalizer.normalize(
            population=population,
        ))
        normalized_record1, normalized_record2 = normalized_population
        assert np.all(np.equal(
            normalized_record1.get_fitness_value(),
            (1, 2, 3),
        ))
        assert normalized_record2.get_fitness_value() is None

    """

    def __init__(
        self,
        coefficient,
        filter=lambda population, record: True,
    ):
        """
        Initialize a :class:`.Multiply` instance.

        Parameters
        ----------
        coefficient : :class:`float` or \
                :class:`tuple` of :class:`float`
            The coefficients each fitness value is multiplied by. Can
            be a single number or multiple numbers, depending on the
            form of the fitness value.

        filter : :class:`callable`, optional
            Takes two parameters, first is a :class:`tuple`
            of :class:`.MoleculeRecord` instances,
            and the second is a :class:`.MoleculeRecord`. The
            :class:`callable` returns ``True`` or ``False``. Only
            molecules which return ``True`` will have fitness values
            normalized. By default, all molecules will have fitness
            values normalized.
            The instance passed to the `population` argument of
            :meth:`.normalize` is passed as the first argument, while
            the second argument will be passed every
            :class:`.MoleculeRecord` in it, one at a time.

        """

        self._coefficient = coefficient
        self._filter = filter

    def normalize(self, population):
        for record in population:
            if self._filter(population, record):
                # Use np.multiply here so that both lists and arrays
                # work properly. If * is used [8]*3 will wrongly make
                # [8, 8, 8].
                yield record.with_fitness_value(
                    fitness_value=np.multiply(
                        self._coefficient,
                        record.get_fitness_value(),
                    ),
                )
            else:
                yield record



"""
Null Fitness Normalizer
=======================

"""

from .fitness_normalizer import FitnessNormalizer


class NullFitnessNormalizer(FitnessNormalizer):
    """
    Does nothing.

    This normalizer just yields the molecule records passed to it,
    without changing them in any way.

    """

    def normalize(self, population):
        yield from population



"""
Twelve Plus Thirty
==================

"""

from scipy.constants import golden

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import LinearVertex, NonLinearVertex


class TwelvePlusThirty(Cage):
    """
    Represents a icosahedron cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )
        bb2 = stk.BuildingBlock(
            smiles='BrC1C(Br)C(Br)C(Br)C1Br',
            functional_groups=[stk.BromoFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.TwelvePlusThirty((bb1, bb2)),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.Collapser` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )
        bb2 = stk.BuildingBlock(
            smiles='BrC1C(Br)C(Br)C(Br)C1Br',
            functional_groups=[stk.BromoFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.TwelvePlusThirty(
                building_blocks=(bb1, bb2),
                optimizer=stk.Collapser(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    Nonlinear building blocks with five functional groups are
    required for this topology.

    Linear building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 5-functional groups: 0 to 11
        | 2-functional groups: 12 to 41

    See :class:`.Cage` for more details and examples.

    """

    # Vertices of a regular origin-centred icosahedron
    # Source: http://eusebeia.dyndns.org/4d/icosahedron
    _non_linears = (
        NonLinearVertex(0, [0, 1, golden]),
        NonLinearVertex(1, [0, -1, golden]),
        NonLinearVertex(2, [0, 1, -golden]),
        NonLinearVertex(3, [0, -1, -golden]),
        NonLinearVertex(4, [1, golden, 0]),
        NonLinearVertex(5, [-1, golden, 0]),
        NonLinearVertex(6, [1, -golden, 0]),
        NonLinearVertex(7, [-1, -golden, 0]),
        NonLinearVertex(8, [golden, 0, 1]),
        NonLinearVertex(9, [-golden, 0, 1]),
        NonLinearVertex(10, [golden, 0, -1]),
        NonLinearVertex(11, [-golden, 0, -1]),
    )

    _vertex_prototypes = (
        *_non_linears,
        LinearVertex.init_at_center(
            id=12,
            vertices=(_non_linears[0], _non_linears[1]),
        ),
        LinearVertex.init_at_center(
            id=13,
            vertices=(_non_linears[0], _non_linears[9]),
        ),
        LinearVertex.init_at_center(
            id=14,
            vertices=(_non_linears[0], _non_linears[5]),
        ),
        LinearVertex.init_at_center(
            id=15,
            vertices=(_non_linears[0], _non_linears[4]),
        ),
        LinearVertex.init_at_center(
            id=16,
            vertices=(_non_linears[0], _non_linears[8]),
        ),
        LinearVertex.init_at_center(
            id=17,
            vertices=(_non_linears[8], _non_linears[1]),
        ),
        LinearVertex.init_at_center(
            id=18,
            vertices=(_non_linears[1], _non_linears[9]),
        ),
        LinearVertex.init_at_center(
            id=19,
            vertices=(_non_linears[9], _non_linears[5]),
        ),
        LinearVertex.init_at_center(
            id=20,
            vertices=(_non_linears[5], _non_linears[4]),
        ),
        LinearVertex.init_at_center(
            id=21,
            vertices=(_non_linears[4], _non_linears[8]),
        ),
        LinearVertex.init_at_center(
            id=22,
            vertices=(_non_linears[5], _non_linears[2]),
        ),
        LinearVertex.init_at_center(
            id=23,
            vertices=(_non_linears[5], _non_linears[11]),
        ),
        LinearVertex.init_at_center(
            id=24,
            vertices=(_non_linears[9], _non_linears[11]),
        ),
        LinearVertex.init_at_center(
            id=25,
            vertices=(_non_linears[9], _non_linears[7]),
        ),
        LinearVertex.init_at_center(
            id=26,
            vertices=(_non_linears[1], _non_linears[7]),
        ),
        LinearVertex.init_at_center(
            id=27,
            vertices=(_non_linears[1], _non_linears[6]),
        ),
        LinearVertex.init_at_center(
            id=28,
            vertices=(_non_linears[8], _non_linears[6]),
        ),
        LinearVertex.init_at_center(
            id=29,
            vertices=(_non_linears[8], _non_linears[10]),
        ),
        LinearVertex.init_at_center(
            id=30,
            vertices=(_non_linears[4], _non_linears[10]),
        ),
        LinearVertex.init_at_center(
            id=31,
            vertices=(_non_linears[4], _non_linears[2]),
        ),
        LinearVertex.init_at_center(
            id=32,
            vertices=(_non_linears[2], _non_linears[11]),
        ),
        LinearVertex.init_at_center(
            id=33,
            vertices=(_non_linears[11], _non_linears[7]),
        ),
        LinearVertex.init_at_center(
            id=34,
            vertices=(_non_linears[7], _non_linears[6]),
        ),
        LinearVertex.init_at_center(
            id=35,
            vertices=(_non_linears[6], _non_linears[10]),
        ),
        LinearVertex.init_at_center(
            id=36,
            vertices=(_non_linears[10], _non_linears[2]),
        ),
        LinearVertex.init_at_center(
            id=37,
            vertices=(_non_linears[2], _non_linears[3]),
        ),
        LinearVertex.init_at_center(
            id=38,
            vertices=(_non_linears[11], _non_linears[3]),
        ),
        LinearVertex.init_at_center(
            id=39,
            vertices=(_non_linears[7], _non_linears[3]),
        ),
        LinearVertex.init_at_center(
            id=40,
            vertices=(_non_linears[6], _non_linears[3]),
        ),
        LinearVertex.init_at_center(
            id=41,
            vertices=(_non_linears[10], _non_linears[3]),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[12], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[12], _vertex_prototypes[1]),
        Edge(2, _vertex_prototypes[13], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[13], _vertex_prototypes[9]),
        Edge(4, _vertex_prototypes[14], _vertex_prototypes[0]),
        Edge(5, _vertex_prototypes[14], _vertex_prototypes[5]),
        Edge(6, _vertex_prototypes[15], _vertex_prototypes[0]),
        Edge(7, _vertex_prototypes[15], _vertex_prototypes[4]),
        Edge(8, _vertex_prototypes[16], _vertex_prototypes[0]),
        Edge(9, _vertex_prototypes[16], _vertex_prototypes[8]),
        Edge(10, _vertex_prototypes[17], _vertex_prototypes[8]),
        Edge(11, _vertex_prototypes[17], _vertex_prototypes[1]),
        Edge(12, _vertex_prototypes[18], _vertex_prototypes[1]),
        Edge(13, _vertex_prototypes[18], _vertex_prototypes[9]),
        Edge(14, _vertex_prototypes[19], _vertex_prototypes[9]),
        Edge(15, _vertex_prototypes[19], _vertex_prototypes[5]),
        Edge(16, _vertex_prototypes[20], _vertex_prototypes[5]),
        Edge(17, _vertex_prototypes[20], _vertex_prototypes[4]),
        Edge(18, _vertex_prototypes[21], _vertex_prototypes[4]),
        Edge(19, _vertex_prototypes[21], _vertex_prototypes[8]),
        Edge(20, _vertex_prototypes[22], _vertex_prototypes[5]),
        Edge(21, _vertex_prototypes[22], _vertex_prototypes[2]),
        Edge(22, _vertex_prototypes[23], _vertex_prototypes[5]),
        Edge(23, _vertex_prototypes[23], _vertex_prototypes[11]),
        Edge(24, _vertex_prototypes[24], _vertex_prototypes[9]),
        Edge(25, _vertex_prototypes[24], _vertex_prototypes[11]),
        Edge(26, _vertex_prototypes[25], _vertex_prototypes[9]),
        Edge(27, _vertex_prototypes[25], _vertex_prototypes[7]),
        Edge(28, _vertex_prototypes[26], _vertex_prototypes[1]),
        Edge(29, _vertex_prototypes[26], _vertex_prototypes[7]),
        Edge(30, _vertex_prototypes[27], _vertex_prototypes[1]),
        Edge(31, _vertex_prototypes[27], _vertex_prototypes[6]),
        Edge(32, _vertex_prototypes[28], _vertex_prototypes[8]),
        Edge(33, _vertex_prototypes[28], _vertex_prototypes[6]),
        Edge(34, _vertex_prototypes[29], _vertex_prototypes[8]),
        Edge(35, _vertex_prototypes[29], _vertex_prototypes[10]),
        Edge(36, _vertex_prototypes[30], _vertex_prototypes[4]),
        Edge(37, _vertex_prototypes[30], _vertex_prototypes[10]),
        Edge(38, _vertex_prototypes[31], _vertex_prototypes[4]),
        Edge(39, _vertex_prototypes[31], _vertex_prototypes[2]),
        Edge(40, _vertex_prototypes[32], _vertex_prototypes[2]),
        Edge(41, _vertex_prototypes[32], _vertex_prototypes[11]),
        Edge(42, _vertex_prototypes[33], _vertex_prototypes[11]),
        Edge(43, _vertex_prototypes[33], _vertex_prototypes[7]),
        Edge(44, _vertex_prototypes[34], _vertex_prototypes[7]),
        Edge(45, _vertex_prototypes[34], _vertex_prototypes[6]),
        Edge(46, _vertex_prototypes[35], _vertex_prototypes[6]),
        Edge(47, _vertex_prototypes[35], _vertex_prototypes[10]),
        Edge(48, _vertex_prototypes[36], _vertex_prototypes[10]),
        Edge(49, _vertex_prototypes[36], _vertex_prototypes[2]),
        Edge(50, _vertex_prototypes[37], _vertex_prototypes[2]),
        Edge(51, _vertex_prototypes[37], _vertex_prototypes[3]),
        Edge(52, _vertex_prototypes[38], _vertex_prototypes[11]),
        Edge(53, _vertex_prototypes[38], _vertex_prototypes[3]),
        Edge(54, _vertex_prototypes[39], _vertex_prototypes[7]),
        Edge(55, _vertex_prototypes[39], _vertex_prototypes[3]),
        Edge(56, _vertex_prototypes[40], _vertex_prototypes[6]),
        Edge(57, _vertex_prototypes[40], _vertex_prototypes[3]),
        Edge(58, _vertex_prototypes[41], _vertex_prototypes[10]),
        Edge(59, _vertex_prototypes[41], _vertex_prototypes[3]),
    )

    _num_windows = 20
    _num_window_types = 1



"""
Sum
===

"""

from .fitness_normalizer import FitnessNormalizer


class Sum(FitnessNormalizer):
    """
    Combines fitness value components into a single fitness value.

    Examples
    --------
    *Combining Fitness Value Components Into a Single Value*

    Sometimes, your :class:`.FitnessCalculator` will return a fitness
    value which is a :class:`tuple` of multiple numbers. Each number
    represents a different property of the molecule, which
    contributes to the final fitness value. This fitness normalizer
    combines these fitness value components into a single number, by
    taking their sum.

    .. testcode:: combining-fitness-value-components

        import stk

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(1, -2, 3),
                normalized=False,
            ),
        )

        # Create the normalizer.
        sum_normalizer = stk.Sum()

        normalized_population = tuple(sum_normalizer.normalize(
            population=population,
        ))
        normalized_record, = normalized_population

        assert normalized_record.get_fitness_value() == 2


    *Selectively Normalizing Fitness Values*

    Sometimes, you only want to normalize some members of a population,
    for example if some do not have an assigned fitness value,
    because the fitness calculation failed for whatever reason.
    You can use the `filter` parameter to exclude records from the
    normalization

    .. testcode:: selectively-normalizing-fitness-values

        import stk

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(1, -2, 3),
                normalized=False,
            ),
            # This will have a fitness value of None.
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ),
        )

        normalizer = stk.Sum(
            # Only normalize values which are not None.
            filter=lambda population, record:
                record.get_fitness_value() is not None,
        )
        normalized_population = tuple(normalizer.normalize(population))
        normalized_record1, normalized_record2 = normalized_population
        assert normalized_record1.get_fitness_value() == 2
        assert normalized_record2.get_fitness_value() is None


    """

    def __init__(self, filter=lambda population, record: True):
        """
        Initialize a :class:`.Sum` instance.

        Parameters
        ----------
        filter : :class:`callable`, optional
            Takes two parameters, first is a :class:`tuple`
            of :class:`.MoleculeRecord` instances,
            and the second is a :class:`.MoleculeRecord`. The
            :class:`callable` returns ``True`` or ``False``. Only
            molecules which return ``True`` will have fitness values
            normalized. By default, all molecules will have fitness
            values normalized.
            The instance passed to the `population` argument of
            :meth:`.normalize` is passed as the first argument, while
            the second argument will be passed every
            :class:`.MoleculeRecord` in it, one at a time.

        """

        self._filter = filter

    def normalize(self, population):
        for record in population:
            if self._filter(population, record):
                yield record.with_fitness_value(
                    fitness_value=sum(record.get_fitness_value()),
                )
            else:
                yield record



"""
Selection Plotter
=================

"""

from collections import Counter
from functools import wraps

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from stk._internal.key_makers.inchi_key import InchiKey

plt.switch_backend("agg")


class SelectionPlotter:
    """
    Plots which molecule records a :class:`.Selector` selects.

    Examples
    --------
    *Plotting Which Molecule Records Got Selected*

    .. testcode:: plotting-which-molecule-records-got-selected

        import stk

        # Make a selector.
        roulette = stk.Roulette(num_batches=10)

        # Make a population.
        population = tuple(
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(i)
            for i in range(100)
        )

        # Make a plotter. You do not have to assign it to a variable.
        stk.SelectionPlotter('roulette_counter', roulette)

        # Select the molecule records.
        selected = tuple(roulette.select(population))

        # There should now be a file called "roulette_counter_1.png"
        # which shows a graph of all the selected records.

        # Select records again.
        selected2 = tuple(roulette.select(population))

        # There should now be a file called "roulette_counter_2.png"
        # which shows a graph of all the selected molecules.

        # And so on every time you use "roulette.select()".

    .. testcode:: plotting-which-molecule-records-got-selected
        :hide:

        import os

        assert os.path.exists('roulette_counter_1.png')
        assert os.path.exists('roulette_counter_1.csv')
        assert os.path.exists('roulette_counter_2.png')
        assert os.path.exists('roulette_counter_2.csv')
        os.remove('roulette_counter_1.png')
        os.remove('roulette_counter_1.csv')
        os.remove('roulette_counter_2.png')
        os.remove('roulette_counter_2.csv')

    """

    def __init__(
        self,
        filename,
        selector,
        x_label="Molecule: InChIKey - Fitness Value",
        record_label=lambda record: (
            f"{InchiKey().get_key(record.get_molecule())} - "
            f"{record.get_fitness_value()}"
        ),
        heat_map_value=lambda record: record.get_fitness_value(),
        heat_map_label="Fitness",
        order_by=lambda record: record.get_fitness_value(),
    ):
        """
        Initialize a :class:`.SelectionPlotter` instance.

        Parameters
        ----------
        filename : :class:`str`
            The basename of the files. This means it should not include
            file extensions.

        selector : :class:`.Selector`
            The :class:`.Selector` whose selection of molecule records
            is plotted.

        x_label : :class:`str`, optional
            The label use for the x axis.

        record_label : :class:`callable`, optional
            A :class:`callable` which takes a :class:`.MoleculeRecord`
            for each record, which is to be
            included on the x-axis of the counter plot. It should
            return a string, which is the label used for the
            :class:`.MoleculeRecord` on the plot.

        heat_map_value : :class:`callable`, optional
            A :class:`callable`, which takes a :class:`.MoleculeRecord`
            for each record, which is to be included on the x-axis, and
            returns a value. The value is used for coloring the heat
            map used in the plot.

        heat_map_label : :class:`str`, optional
            The label used for the heat map key.

        order_by : :class:`callable`, optional
            A :class:`callable`, which takes a :class:`.MoleculeRecord`
            for each record, which is to be included on the x-axis,
            and returns a value. The value is used to sort the plotted
            records along the x-axis in descending order.

        """

        self._plots = 0
        self._filename = filename
        self._x_label = x_label
        self._record_label = record_label
        self._order_by = order_by
        self._heat_map_value = heat_map_value
        self._heat_map_label = heat_map_label
        selector.select = self._update_counter(selector.select)

    def _update_counter(self, select):
        """
        Decorate :meth:`.Selector.select`.

        This is a decorator which makes sure that every time
        :meth:`.Selector.select` selects a :class:`.MoleculeRecord` a
        counter keeping track of selected records is updated.

        Parameters
        ----------
        select : :class:`callable`
            The :meth:`Selector.select` method to decorate.

        Returns
        -------
        :class:`function`
            The decorated :meth:`.Selector.select` method.

        """

        @wraps(select)
        def inner(population, *args, **kwargs):
            counter = Counter({record: 0 for record in population})
            for selected in select(population, *args, **kwargs):
                counter.update(selected)
                yield selected
            self._plot(population, counter)

        return inner

    def _plot(self, population, counter):
        """
        Plot a selection counter.

        Parameters
        ----------
        population : :class:`tuple` of :class:`.MoleculeRecord`
            The population from which molecule records were selected.

        counter : :class:`collections.Counter`
            A counter specifying which records were selected and how
            many times.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._plots += 1
        sns.set(style="darkgrid")
        data = []
        for record, selection_count in counter.items():
            label = self._record_label(record)
            data.append(
                pd.DataFrame(
                    data={
                        self._x_label: label,
                        "Number of Times Selected": selection_count,
                        "order": self._order_by(record),
                        "heat_map": self._heat_map_value(record),
                    },
                    index=[self._x_label],
                )
            )
        df = pd.concat(data, ignore_index=True)

        df = df.sort_values(
            ["Number of Times Selected", "order"],
            ascending=[False, False],
        )
        norm = plt.Normalize(df["heat_map"].min(), df["heat_map"].max())
        sm = plt.cm.ScalarMappable(cmap="magma_r", norm=norm)
        sm.set_array([])

        df.to_csv(f"{self._filename}_{self._plots}.csv")
        fig, ax = plt.subplots(figsize=(11.7, 8.28))
        sns.scatterplot(
            x="Number of Times Selected",
            y=self._x_label,
            hue="heat_map",
            palette="magma_r",
            data=df,
            s=[200 for i in range(len(counter.keys()))],
            ax=ax,
        )
        ax.get_legend().remove()
        # https://tinyurl.com/2p9drmkh
        plt.rcParams["axes.grid"] = False
        ax.figure.colorbar(sm, ax=ax).set_label(self._heat_map_label)
        plt.tight_layout()
        fig.savefig(f"{self._filename}_{self._plots}.png", dpi=fig.dpi)
        plt.close("all")



import typing

from stk._internal.constructed_molecule import ConstructedMolecule
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)


class MoleculeRecord:
    """
    An abstract base class for molecular records used by the EA.

    Notes:

        You might notice that the public methods of this abstract base
        class are implemented. This is a default implementation provided
        purely for convenience. Subclasses can freely ignore or
        override this implementation.

    """

    def __init__(self, topology_graph: TopologyGraph) -> None:
        """
        Parameters:
            topology_graph:
                The topology graph of a :class:`.ConstructedMolecule`.
        """
        self._molecule = ConstructedMolecule(topology_graph)
        self._topology_graph = topology_graph
        self._fitness_value = None
        self._normalized_fitness_value = None

    def get_molecule(self) -> ConstructedMolecule:
        """
        Get the molecule held by the record.

        Returns:
            The molecule held by the record.

        """

        return self._molecule

    def get_topology_graph(self) -> TopologyGraph:
        """
        Get the topology graph of the molecule.

        Returns:
            The topology graph.

        """

        return self._topology_graph

    def get_fitness_value(
        self, normalized: bool = True
    ) -> typing.Any | float | None:
        """
        Get the fitness value of the molecule in the record.

        Parameters:

            normalized:
                Toggles the return of the normalized vs unnormalized
                fitness value. The unnormalized fitness value is
                guaranteed to be constant for the same molecule
                across generations, while the normalized one is allowed
                to change.

        Returns:
            The fitness value.
        """

        return (
            self._normalized_fitness_value
            if normalized
            else self._fitness_value
        )

    def clone(self) -> typing.Self:
        """
        Return a clone.

        Returns:
            MoleculeRecord: The clone. Has the same type as
            the original record.

        """

        clone = self.__class__.__new__(self.__class__)
        clone._molecule = self._molecule
        clone._topology_graph = self._topology_graph
        clone._fitness_value = self._fitness_value
        clone._normalized_fitness_value = self._normalized_fitness_value
        return clone

    def with_fitness_value(
        self,
        fitness_value: typing.Any,
        normalized: bool = True,
    ) -> typing.Self:
        """
        Return a clone holding a different fitness value.

        Parameters:
            fitness_value:
                The fitness value of the clone.

            normalized:
                Toggles if the normalized or unnormalized fitness value is
                being set. If ``False``, both the normalized and
                unnormalized fitness values with be set to `fitness_value`.

        Returns:
            MoleculeRecord: The clone. Has the same type as
            the original record.

        """

        return self.clone()._with_fitness_value(
            fitness_value=fitness_value,
            normalized=normalized,
        )

    def _with_fitness_value(self, fitness_value, normalized):
        if not normalized:
            self._fitness_value = fitness_value
        self._normalized_fitness_value = fitness_value
        return self



"""
Add
===

"""

import numpy as np

from .fitness_normalizer import FitnessNormalizer


class Add(FitnessNormalizer):
    """
    Adds a number to the fitness values.

    Examples
    --------
    *Incrementing Fitness Values by a Set of Values*

    In this example, assume that each fitness value consists of a
    :class:`tuple` of numbers, each representing a different property
    of the molecule, and each contributing to the final fitness value.
    The properties can be anything, such as  energy, number of atoms
    or diameter.

    Often, if these properties indicate a low fitness value, you
    will take their inverse. However, if these properties can have
    a value of 0, and you try to take their inverse you can end up
    dividing by 0, which is bad. To avoid this, you can add a number,
    like 1, to the fitness values before taking their inverse. This
    normalizer allows you to do this.

    Giving a concrete example

    .. testcode:: incrementing-fitness-values-by-a-set-of-values

        import stk
        import numpy as np

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(0, 0, 0),
                normalized=False,
            ),
        )

        normalizer = stk.Add((1, 2, 3))
        # Calling normalizer.normalize() will return a new
        # population holding the molecule records with normalized
        # fitness values.
        normalized_population = tuple(normalizer.normalize(
            population=population,
        ))
        normalized_record, = normalized_population
        assert np.all(np.equal(
            normalized_record.get_fitness_value(),
            (1, 2, 3),
        ))

    *Selectively Normalizing Fitness Values*

    Sometimes, you only want to normalize some members of a population,
    for example if some do not have an assigned fitness value,
    because the fitness calculation failed for whatever reason.
    You can use the `filter` parameter to exclude records from the
    normalization

    .. testcode:: incrementing-fitness-values-by-a-set-of-values

        import stk
        import numpy as np

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(0, 0, 0),
                normalized=False,
            ),
            # This will have a fitness value of None.
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ),
        )

        normalizer = stk.Add(
            number=(1, 2, 3),
            # Only normalize values which are not None.
            filter=lambda population, record:
                record.get_fitness_value() is not None,
        )

        # Calling normalizer.normalize() will return a new
        # population holding the molecule records with normalized
        # fitness values.
        normalized_population = tuple(normalizer.normalize(
            population=population,
        ))
        normalized_record1, normalized_record2 = normalized_population
        assert np.all(np.equal(
            normalized_record1.get_fitness_value(),
            (1, 2, 3),
        ))
        assert normalized_record2.get_fitness_value() is None

    """

    def __init__(
        self,
        number,
        filter=lambda population, record: True,
    ):
        """
        Initialize a :class:`.Add` instance.

        Parameters
        ----------
        number : :class:`float` or \
                :class:`tuple` of :class:`float`
            The number each fitness value is increased by. Can
            be a single number or multiple numbers, depending on the
            form of the fitness value.

        filter : :class:`callable`, optional
            Takes two parameters, first is a :class:`tuple`
            of :class:`.MoleculeRecord` instances,
            and the second is a :class:`.MoleculeRecord`. The
            :class:`callable` returns ``True`` or ``False``. Only
            molecules which return ``True`` will have fitness values
            normalized. By default, all molecules will have fitness
            values normalized.
            The instance passed to the `population` argument of
            :meth:`.normalize` is passed as the first argument, while
            the second argument will be passed every
            :class:`.MoleculeRecord` in it, one at a time.

        """

        self._number = number
        self._filter = filter

    def normalize(self, population):
        for record in population:
            if self._filter(population, record):
                # Use np.add here so that both tuples and arrays
                # work properly.
                yield record.with_fitness_value(
                    fitness_value=np.add(
                        self._number,
                        record.get_fitness_value(),
                    ),
                )
            else:
                yield record



"""
M6L12 Cube
==========

"""

import numpy as np

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import LinearVertex, NonLinearVertex


class M6L12Cube(Cage):
    """
    Represents a cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(4)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=NC=CC(C2=CC=CC(C3=C'
                'C=NC=C3)=C2)=C1'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M6L12Cube(
                building_blocks=(bb1, bb2),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.Collapser` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(4)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=NC=CC(C2=CC=CC(C3=C'
                'C=NC=C3)=C2)=C1'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M6L12Cube(
                building_blocks=(bb1, bb2),
                optimizer=stk.Collapser(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    Metal building blocks with four functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 4-functional groups: 0 to 5
        | 2-functional groups: 6 to 17

    See :class:`.Cage` for more details and examples.

    """

    _x = np.sqrt(2)
    _vertex_prototypes = (
        NonLinearVertex(0, [_x, 0, 0]),
        NonLinearVertex(1, [0, _x, 0]),
        NonLinearVertex(2, [-_x, 0, 0]),
        NonLinearVertex(3, [0, -_x, 0]),
        NonLinearVertex(4, [0, 0, _x]),
        NonLinearVertex(5, [0, 0, -_x]),
        LinearVertex(6, [1, 1, 0], False),
        LinearVertex(7, [1, -1, 0], False),
        LinearVertex(8, [1, 0, 1], False),
        LinearVertex(9, [1, 0, -1], False),
        LinearVertex(10, [-1, 1, 0], False),
        LinearVertex(11, [-1, -1, 0], False),
        LinearVertex(12, [-1, 0, 1], False),
        LinearVertex(13, [-1, 0, -1], False),
        LinearVertex(14, [0, 1, 1], False),
        LinearVertex(15, [0, 1, -1], False),
        LinearVertex(16, [0, -1, 1], False),
        LinearVertex(17, [0, -1, -1], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[6]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[7]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[8]),
        Edge(3, _vertex_prototypes[0], _vertex_prototypes[9]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[6]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[10]),
        Edge(6, _vertex_prototypes[1], _vertex_prototypes[14]),
        Edge(7, _vertex_prototypes[1], _vertex_prototypes[15]),
        Edge(8, _vertex_prototypes[2], _vertex_prototypes[10]),
        Edge(9, _vertex_prototypes[2], _vertex_prototypes[11]),
        Edge(10, _vertex_prototypes[2], _vertex_prototypes[12]),
        Edge(11, _vertex_prototypes[2], _vertex_prototypes[13]),
        Edge(12, _vertex_prototypes[3], _vertex_prototypes[7]),
        Edge(13, _vertex_prototypes[3], _vertex_prototypes[11]),
        Edge(14, _vertex_prototypes[3], _vertex_prototypes[16]),
        Edge(15, _vertex_prototypes[3], _vertex_prototypes[17]),
        Edge(16, _vertex_prototypes[4], _vertex_prototypes[8]),
        Edge(17, _vertex_prototypes[4], _vertex_prototypes[12]),
        Edge(18, _vertex_prototypes[4], _vertex_prototypes[14]),
        Edge(19, _vertex_prototypes[4], _vertex_prototypes[16]),
        Edge(20, _vertex_prototypes[5], _vertex_prototypes[9]),
        Edge(21, _vertex_prototypes[5], _vertex_prototypes[13]),
        Edge(22, _vertex_prototypes[5], _vertex_prototypes[15]),
        Edge(23, _vertex_prototypes[5], _vertex_prototypes[17]),
    )

    _num_windows = 8
    _num_window_types = 1



import numpy as np

from stk._internal.constructed_molecule import ConstructedMolecule
from stk._internal.key_makers.inchi_key import InchiKey
from stk._internal.key_makers.molecule import MoleculeKeyMaker

from .molecule import MoleculeDejsonizer, MoleculeJsonizer
from .utilities import to_atom, to_atom_info, to_bond, to_bond_info


class ConstructedMoleculeJsonizer:
    """
    Abstract base class for creating JSONs of constructed molecules.

    See Also
    --------
    :class:`.MoleculeJsonizer`

    Notes
    -----
    You might notice that the public methods of this abstract base
    class are implemented. These are just default implementations,
    which can be safely ignored or overridden, when implementing
    subclasses. However, the default implementation can be used
    directly, if it suits your needs.

    Examples
    --------
    *Converting a Constructed Molecule to JSON*

    You want get a JSON representation of a
    :class:`.ConstructedMolecule`

    .. testcode:: converting-a-molecule-to-json

        import stk

        # Make the molecule you want jsonize.
        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(
                    stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                ),
                repeating_unit='A',
                num_repeating_units=3,
            )
        )

        # Make a JSONizer.
        jsonizer = stk.ConstructedMoleculeJsonizer()
        # Get the JSON.
        json = jsonizer.to_json(polymer)

    *Adding Additional Molecular Keys*

    Apart from atoms, bonds and the position matrix, the JSON
    representation holds additional fields, one for each
    :class:`.MoleculeKeyMaker` provided to the initializer

    .. testcode:: adding-additional-molecular-keys

        import stk

        # Make the molecule you want jsonize.
        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(
                    stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                ),
                repeating_unit='A',
                num_repeating_units=3,
            )
        )

        # Make a JSONizer.
        jsonizer = stk.ConstructedMoleculeJsonizer()
        # Get the JSON.
        json = jsonizer.to_json(polymer)

    In this case, ``json`` will look something like

    .. code-block:: python

        {
            # A tuple of JSON atom representations.
            'atoms': (...),

            # A tuple of JSON bond representations.
            'bonds': (...),

            'InChI': 'The InChI of the molecule',
            'InChIKey': 'The InChIKey of the molecule',
        }

    For every :class:`.MoleculeKeyMaker` provided to `key_makers`,
    a new key will be added to the JSON representation, with its name
    given by :meth:`.MoleculeKeyMaker.get_key_name` and the value
    given by :meth:`.MoleculeKeyMaker.get_key`.

    """

    def __init__(
        self,
        key_makers=(InchiKey(),),
    ):
        """
        Initializes a :class:`.ConstructedMoleculeJsonizer`.

        Parameters
        ----------
        key_makers : :class:`tuple` of :class:`.MoleculeKeyMaker`
            Used to make the keys of molecules, which should be
            included in their JSON representations. Keys allow
            molecular data to reference itself when split across
            multiple JSONs.

        """

        self._jsonizer = MoleculeJsonizer(key_makers=())
        self._key_makers = key_makers

    def to_json(self, molecule):
        """
        Serialize `molecule`.

        Parameters
        ----------
        molecule : :class:`.ConstructedMolecule`
            The constructed molecule to serialize.

        Returns
        -------
        :class:`dict`
            A JSON representation of `molecule`.

        """

        def get_keys(building_block):
            return {
                key_maker.get_key_name(): key_maker.get_key(building_block)
                for key_maker in self._key_makers
                if isinstance(key_maker, MoleculeKeyMaker)
            }

        building_block_indices = {
            building_block: index
            for index, building_block in enumerate(
                molecule.get_building_blocks()
            )
        }
        building_block_indices[None] = None

        def atom_info_to_json(atom_info):
            if atom_info.get_building_block() is None:
                return (
                    None,
                    None,
                    None,
                )

            return (
                building_block_indices[atom_info.get_building_block()],
                atom_info.get_building_block_id(),
                atom_info.get_building_block_atom().get_id(),
            )

        def bond_info_to_json(bond_info):
            return (
                building_block_indices[bond_info.get_building_block()],
                bond_info.get_building_block_id(),
            )

        molecule_json = self._jsonizer.to_json(molecule)
        constructed_molecule_json = {
            "BB": tuple(
                map(
                    get_keys,
                    molecule.get_building_blocks(),
                )
            ),
            "aI": tuple(
                map(
                    atom_info_to_json,
                    molecule.get_atom_infos(),
                )
            ),
            "bI": tuple(
                map(
                    bond_info_to_json,
                    molecule.get_bond_infos(),
                )
            ),
            "nBB": tuple(
                map(
                    molecule.get_num_building_block,
                    molecule.get_building_blocks(),
                )
            ),
        }
        for key_maker in self._key_makers:
            key_name = key_maker.get_key_name()
            key = key_maker.get_key(molecule)
            molecule_json["molecule"][key_name] = key
            molecule_json["matrix"][key_name] = key
            constructed_molecule_json[key_name] = key

        building_block_jsons = tuple(
            map(
                self._jsonizer.to_json,
                molecule.get_building_blocks(),
            )
        )

        def is_molecule_key_maker(key_maker):
            return isinstance(key_maker, MoleculeKeyMaker)

        for key_maker in filter(
            is_molecule_key_maker,
            self._key_makers,
        ):
            key_name = key_maker.get_key_name()
            for building_block, json in zip(
                molecule.get_building_blocks(),
                building_block_jsons,
            ):
                key = key_maker.get_key(building_block)
                json["molecule"][key_name] = key
                json["matrix"][key_name] = key

        return {
            "molecule": molecule_json["molecule"],
            "constructedMolecule": constructed_molecule_json,
            "matrix": molecule_json["matrix"],
            "buildingBlocks": building_block_jsons,
        }

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"{self.__class__.__name__}({self._key_makers!r})"


class ConstructedMoleculeDejsonizer:
    """
    Abstract base class for creating constructed molecules from JSONs.

    See Also
    --------
    :class:`.MoleculeDejsonizer`

    Notes
    -----
    You might notice that the public methods of this abstract base
    class are implemented. These are just default implementations,
    which can be safely ignored or overridden, when implementing
    subclasses. However, the default implementation can be used
    directly, if it suits your needs.

    """

    def __init__(self):
        """
        Initialize a :class:`.ConstructedMoleculeDejsonizer` instance.

        """

        self._dejsonizer = MoleculeDejsonizer()

    def from_json(self, json):
        """
        Get a :class:`.ConstructedMolecule` from a JSON.

        Parameters
        ----------
        json : :class:`dict`
            A JSON of the constructed molecule.

        Returns
        -------
        :class:`.ConstructedMolecule`
            The constructed molecule.

        """

        building_blocks = tuple(
            map(
                self._dejsonizer.from_json,
                json["buildingBlocks"],
            )
        )

        num_building_blocks = (
            (building_block, num)
            for building_block, num in zip(
                building_blocks,
                json["constructedMolecule"]["nBB"],
            )
        )
        atoms = tuple(
            to_atom(atom_id, atom_json)
            for atom_id, atom_json in enumerate(json["molecule"]["a"])
        )
        bonds = tuple(
            to_bond(atoms, bond_json) for bond_json in json["molecule"]["b"]
        )
        return ConstructedMolecule.init(
            atoms=atoms,
            bonds=bonds,
            position_matrix=np.array(json["matrix"]["m"]),
            atom_infos=tuple(
                to_atom_info(
                    building_blocks=building_blocks,
                    atom=atoms[atom_id],
                    json=atom_info_json,
                )
                for atom_id, atom_info_json in enumerate(
                    json["constructedMolecule"]["aI"]
                )
            ),
            bond_infos=tuple(
                to_bond_info(
                    building_blocks=building_blocks,
                    bond=bonds[bond_id],
                    json=bond_info_json,
                )
                for bond_id, bond_info_json in enumerate(
                    json["constructedMolecule"]["bI"]
                )
            ),
            num_building_blocks=num_building_blocks,
        )

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"{self.__class__.__name__}()"



"""
Constructed Molecule
====================

"""

from __future__ import annotations

import logging
import typing

import numpy as np
import rdkit.Chem.AllChem as rdkit

from stk._internal.atom import Atom
from stk._internal.atom_info import AtomInfo
from stk._internal.bond import Bond
from stk._internal.bond_info import BondInfo
from stk._internal.construction_result.construction_result import (
    ConstructionResult,
)
from stk._internal.molecule import Molecule
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.utilities.molecule import (
    get_bond_info_atom_ids,
    sort_bond_atoms_by_id,
)
from stk._internal.utilities.utilities import OneOrMany

logger = logging.getLogger(__name__)


class ConstructedMolecule(Molecule):
    """
    Represents constructed molecules.

    Examples:

        *Initialization*

        A :class:`.ConstructedMolecule` is initialized from a
        :class:`.TopologyGraph`, which is typically initialized from
        some :class:`.BuildingBlock` instances.

        .. testcode:: initialization

            import stk

            bb1 = stk.BuildingBlock(
                smiles='NCCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(C=O)CC=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            tetrahedron = stk.cage.FourPlusSix((bb1, bb2))
            cage = stk.ConstructedMolecule(tetrahedron)

        *Hierarchical Construction*

        A :class:`ConstructedMolecule` may be used to construct other
        :class:`ConstructedMolecule` instances, though you will
        probably have to convert it to a :class:`.BuildingBlock` first

        .. testcode:: hierarchical-construction

            import stk

            bb1 = stk.BuildingBlock(
                smiles='NCCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(C=O)CC=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            tetrahedron = stk.cage.FourPlusSix((bb1, bb2))
            cage = stk.ConstructedMolecule(tetrahedron)

            benzene = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('c1ccccc1'),
            )
            cage_complex = stk.host_guest.Complex(
                host=stk.BuildingBlock.init_from_molecule(cage),
                guests=benzene,
            )
            cage_complex = stk.ConstructedMolecule(cage_complex)

        Obviously, the initialization of the
        :class:`.ConstructedMolecule` depends mostly on the specifics
        of the :class:`.TopologyGraph` used, and the documentation of
        those classes should be examined for more examples.

    """

    _atom_infos: tuple[AtomInfo, ...]
    _bond_infos: tuple[BondInfo, ...]
    _num_building_blocks: dict[Molecule, int]

    def __init__(
        self,
        topology_graph: TopologyGraph,
    ) -> None:
        """
        Initialize a :class:`.ConstructedMolecule`.

        Parameters:

            topology_graph:
                The topology graph of the constructed molecule.

        """

        self._init_from_construction_result(
            obj=self,
            construction_result=topology_graph.construct(),
        )

    @classmethod
    def init(
        cls,
        atoms: tuple[Atom, ...],
        bonds: tuple[Bond, ...],
        position_matrix: np.ndarray,
        atom_infos: tuple[AtomInfo, ...],
        bond_infos: tuple[BondInfo, ...],
        num_building_blocks: dict[Molecule, int],
    ) -> ConstructedMolecule:
        """
        Initialize a :class:`.ConstructedMolecule` from its components.

        Parameters:

            atoms:
                The atoms of the molecule.

            bond:
                The bonds of the molecule.

            position_matrix:
                A ``(n, 3)`` position matrix of the molecule.

            atom_infos:
                The atom infos of the molecule.

            bond_infos:
                The bond infos of the molecule.

            num_building_blocks:
                Maps each building block of the constructed molecule to
                the number of times it is present in it.

        Returns:

            The constructed molecule.

        """

        molecule = cls.__new__(cls)
        Molecule.__init__(molecule, atoms, bonds, position_matrix)
        molecule._atom_infos = atom_infos
        molecule._bond_infos = bond_infos
        molecule._num_building_blocks = dict(num_building_blocks)
        return molecule

    @classmethod
    def init_from_construction_result(
        cls,
        construction_result: ConstructionResult,
    ) -> ConstructedMolecule:
        """
        Initialize a :class:`.ConstructedMolecule`.

        Parameters:

            construction_result:
                The result of a construction, from which the
                :class:`.ConstructedMolecule` should be initialized.

        Returns:

            The constructed molecule.

        """

        return cls._init_from_construction_result(
            obj=cls.__new__(cls),
            construction_result=construction_result,
        )

    @staticmethod
    def _init_from_construction_result(
        obj: ConstructedMolecule,
        construction_result: ConstructionResult,
    ) -> ConstructedMolecule:
        """
        Initialize a :class:`.ConstructedMolecule`.

        This modifies `obj`.

        Parameters:

            obj:
                The constructed molecule to initialize.

            construction_result:
                The result of a construction, from which the
                :class:`.ConstructedMolecule` should be initialized.

        Returns:

            The `obj` instance.

        """

        super(ConstructedMolecule, obj).__init__(
            atoms=construction_result.get_atoms(),
            bonds=construction_result.get_bonds(),
            position_matrix=construction_result.get_position_matrix(),
        )
        obj._atom_infos = construction_result.get_atom_infos()
        obj._bond_infos = construction_result.get_bond_infos()
        obj._num_building_blocks = {
            building_block: construction_result.get_num_building_block(
                building_block=building_block,
            )
            for building_block in construction_result.get_building_blocks()
        }
        return obj

    def clone(self) -> ConstructedMolecule:
        clone = self._clone()
        clone._atom_infos = self._atom_infos
        clone._bond_infos = self._bond_infos
        clone._num_building_blocks = dict(self._num_building_blocks)
        return clone

    def get_building_blocks(self) -> typing.Iterator[Molecule]:
        """
        Yield the building blocks of the constructed molecule.

        Building blocks are yielded in an order based on their
        position in the constructed molecule. For two topologically
        equivalent constructed molecules, but with different building
        blocks, equivalently positioned building blocks will be
        yielded at the same time.

        Yields:

            A building block of the constructed molecule.

        """

        yield from self._num_building_blocks

    def get_num_building_block(
        self,
        building_block: Molecule,
    ) -> int:
        """
        Get the number of times `building_block` is present.

        Parameters:

            building_block:
                The building block whose frequency in the constructed
                molecule is desired.

        Returns:

            The number of times `building_block` was used in the
            construction of the constructed molecule.

        """

        return self._num_building_blocks[building_block]

    def get_atom_infos(
        self,
        atom_ids: typing.Optional[OneOrMany[int]] = None,
    ) -> typing.Iterator[AtomInfo]:
        """
        Yield data about atoms in the molecule.

        Parameters:

            atom_ids:
                The ids of atoms whose data is desired. If ``None``,
                data on all atoms will be yielded. Can be a single
                :class:`int`, if data on a single atom is desired.

        Yields:

            Data about an atom.

        """

        if atom_ids is None:
            atom_ids = range(len(self._atoms))
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids,)

        for atom_id in atom_ids:
            yield self._atom_infos[atom_id]

    def get_bond_infos(self) -> typing.Iterator[BondInfo]:
        """
        Yield data about bonds in the molecule.

        Yields:

            Data about a bond.

        """

        yield from self._bond_infos

    def with_canonical_atom_ordering(self) -> ConstructedMolecule:
        return self.clone()._with_canonical_atom_ordering()

    def _with_canonical_atom_ordering(self) -> ConstructedMolecule:
        # Make all building blocks canonically ordered too.
        building_blocks = {
            building_block: building_block.with_canonical_atom_ordering()
            for building_block in self._num_building_blocks
        }

        # Cache these mappings for later, to avoid unnecessary
        # re-computations of canonical ordering.
        canonical_map = {
            building_block: building_block.get_canonical_atom_ids()
            for building_block in self._num_building_blocks
        }

        self._num_building_blocks = dict(
            zip(
                building_blocks.values(),
                self._num_building_blocks.values(),
            ),
        )

        ordering = rdkit.CanonicalRankAtoms(self.to_rdkit_mol())
        id_map = {
            new_id: atom.get_id()
            for new_id, atom in zip(ordering, self._atoms)
        }
        super()._with_canonical_atom_ordering()
        atom_map = {
            old_id: self._atoms[new_id]
            for old_id, new_id in enumerate(ordering)
        }
        old_atom_infos = self._atom_infos

        def get_atom_info(atom: Atom) -> AtomInfo:
            old_atom_info = old_atom_infos[id_map[atom.get_id()]]
            old_building_block = old_atom_info.get_building_block()

            if old_building_block is None:
                return AtomInfo(
                    atom=atom,
                    building_block_atom=None,
                    building_block=None,
                    building_block_id=None,
                )

            # If old_building_block is not None then neither is the
            # old_building_block_atom.
            old_building_block_atom = typing.cast(
                Atom,
                old_atom_info.get_building_block_atom(),
            )

            canonical_building_block_atom_id = canonical_map[
                old_building_block
            ][old_building_block_atom.get_id()]

            canonical_building_block = building_blocks[old_building_block]

            (
                canonical_building_block_atom,
            ) = canonical_building_block.get_atoms(
                atom_ids=canonical_building_block_atom_id,
            )

            return AtomInfo(
                atom=atom,
                building_block_atom=canonical_building_block_atom,
                building_block=canonical_building_block,
                building_block_id=(old_atom_info.get_building_block_id()),
            )

        def get_bond_info(info: BondInfo) -> BondInfo:
            building_block = info.get_building_block()
            return BondInfo(
                bond=sort_bond_atoms_by_id(
                    info.get_bond().with_atoms(atom_map)
                ),
                building_block=(
                    building_block
                    if building_block is None
                    else building_blocks[building_block]
                ),
                building_block_id=info.get_building_block_id(),
            )

        self._atom_infos = tuple(map(get_atom_info, self._atoms))
        self._bond_infos = tuple(
            sorted(
                map(get_bond_info, self._bond_infos),
                key=get_bond_info_atom_ids,
            )
        )
        return self

    def with_centroid(
        self,
        position: np.ndarray,
        atom_ids: typing.Optional[OneOrMany[int]] = None,
    ) -> ConstructedMolecule:
        return self.clone()._with_centroid(position, atom_ids)

    def with_displacement(
        self,
        displacement: np.ndarray,
    ) -> ConstructedMolecule:
        return self.clone()._with_displacement(displacement)

    def with_position_matrix(
        self,
        position_matrix: np.ndarray,
    ) -> ConstructedMolecule:
        return self.clone()._with_position_matrix(position_matrix)

    def with_rotation_about_axis(
        self,
        angle: float,
        axis: np.ndarray,
        origin: np.ndarray,
    ) -> ConstructedMolecule:
        return self.clone()._with_rotation_about_axis(
            angle=angle,
            axis=axis,
            origin=origin,
        )

    def with_rotation_between_vectors(
        self,
        start: np.ndarray,
        target: np.ndarray,
        origin: np.ndarray,
    ) -> ConstructedMolecule:
        return self.clone()._with_rotation_between_vectors(
            start=start,
            target=target,
            origin=origin,
        )

    def with_rotation_to_minimize_angle(
        self,
        start: np.ndarray,
        target: np.ndarray,
        axis: np.ndarray,
        origin: np.ndarray,
    ) -> ConstructedMolecule:
        return self.clone()._with_rotation_to_minimize_angle(
            start=start,
            target=target,
            axis=axis,
            origin=origin,
        )

    def with_structure_from_file(
        self,
        path: str,
        extension: typing.Optional[str] = None,
    ) -> ConstructedMolecule:
        return typing.cast(
            ConstructedMolecule,
            super().with_structure_from_file(path, extension),
        )

    def write(
        self,
        path: str,
        atom_ids: typing.Optional[OneOrMany[int]] = None,
    ) -> ConstructedMolecule:
        return typing.cast(
            ConstructedMolecule,
            super().write(path, atom_ids),
        )



"""
Divide By Mean
==============

"""

import logging
from functools import partial

import numpy as np

from .fitness_normalizer import FitnessNormalizer

logger = logging.getLogger(__name__)


class DivideByMean(FitnessNormalizer):
    """
    Divides fitness values by the population mean.

    While this function can be used if the fitness value of each
    :class:`.Molecule` in the population is a single
    number, it is most useful when the fitness value is a
    :class:`tuple` of numbers. In this case, it is necessary to somehow
    combine the numbers so that a single fitness value is produced.
    For example, take a fitness value which is the vector holding the
    properties ``[energy, diameter, num_atoms]``. For a given molecule
    these numbers may be something like ``[200,000, 12, 140]``. If we
    were to sum these numbers, the energy term would dominate the final
    fitness value. In order to combine these numbers we can divide them
    by the population averages. For example, if the average energy
    of molecules in the population is ``300,000`` the average diameter
    is ``10`` and the average number of atoms is ``70`` then the
    fitness vector would be scaled to ``[0.5, 1.2, 2]``. These
    numbers are now of a similar magnitude and can be summed to give a
    reasonable value. After division , each value represents how
    much better than the population average each property value is.
    In essence we have removed the units from each parameter.

    Examples
    --------
    *Selectively Normalizing Fitness Values*

    Sometimes you do not want to normalize all the values in a
    population together. For example, if a failed fitness value
    calculation resulted in some records having a fitness value of
    ``None``, you would want to ignore these records from the
    normalization

    .. testcode:: selectively-normalizing-fitness-values

        import stk
        import numpy as np

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(1., 2., 3.),
                normalized=False,
            ),
            # This will have a fitness value of None.
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ),
        )

        mean_scaler = stk.DivideByMean(
            # Only normalize values which are not None.
            filter=lambda population, record:
                record.get_fitness_value() is not None
        )
        # Calling mean_scaler.normalize() will return a new
        # population holding the molecule records with normalized
        # fitness values.
        normalized_population = tuple(mean_scaler.normalize(
            population=population,
        ))
        normalized_record1, normalized_record2 = normalized_population
        assert np.all(np.equal(
            normalized_record1.get_fitness_value(),
            (1, 1, 1),
        ))

    """

    def __init__(self, filter=lambda population, record: True):
        """
        Initialize a :class:`.DivideByMean` instance.

        Parameters
        ----------
        filter : :class:`callable`, optional
            Takes two parameters, first is a :class:`tuple`
            of :class:`.MoleculeRecord` instances,
            and the second is a :class:`.MoleculeRecord`. The
            :class:`callable` returns ``True`` or ``False``. Only
            molecules which return ``True`` will have fitness values
            normalized. By default, all molecules will have fitness
            values normalized.
            The instance passed to the `population` argument of
            :meth:`.normalize` is passed as the first argument, while
            the second argument will be passed every
            :class:`.MoleculeRecord` in it, one at a time.

        """

        self._filter = filter

    def normalize(self, population):
        filtered = filter(
            partial(self._filter, population),
            population,
        )
        mean = np.mean(
            a=[record.get_fitness_value() for record in filtered],
            axis=0,
        )
        logger.debug(f"Means used: {mean}")

        for record in population:
            if self._filter(population, record):
                yield record.with_fitness_value(
                    fitness_value=np.divide(
                        record.get_fitness_value(),
                        mean,
                    )
                )
            else:
                yield record



"""
Filter Batches
==============

"""

from .selector import Selector


class FilterBatches(Selector):
    """
    Allows a :class:`.Selector` to select only some batches.

    Examples
    --------
    *Using a Selection Algorithm on a Subset of Batches*

    You only want the :class:`.Best` 10 batches to participate in
    :class:`.Roulette`

    .. testcode:: using-a-selection-algorithm-on-a-subset-of-batches

        import stk

        # Select the 10 best batches first, and then run roulette on
        # those.
        selector = stk.FilterBatches(
            filter=stk.Best(10),
            selector=stk.Roulette(7),
        )

        population = tuple(
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(i)
            for i in range(100)
        )

        for batch in selector.select(population):
            # Do stuff with batch. It is one of the 10 best batches and
            # was selected using roulette selection.
            pass

    """

    def __init__(self, filter, selector):
        """
        Initialize a :class:`.FilterBatches` instance.

        Parameters
        ----------
        filter : :class:`.Selector`
            Selects batches which can be yielded by `selector`.

        selector : :class:`.Selector`
            Selects batches, but only if they were also selected by
            `filter`.

        """

        self._filter = filter
        self._selector = selector

    def select(
        self,
        population,
        included_batches=None,
        excluded_batches=None,
    ):
        allowed = {
            batch.get_identity_key()
            for batch in self._filter.select(
                population=population,
                included_batches=included_batches,
                excluded_batches=excluded_batches,
            )
        }
        yield from self._selector.select(
            population=population,
            included_batches=allowed,
            excluded_batches=excluded_batches,
        )



"""
Shift Up
========

"""

from functools import partial

import numpy as np

from .fitness_normalizer import FitnessNormalizer


class ShiftUp(FitnessNormalizer):
    """
    Shifts negative fitness values to be positive.

    Assume you have a vector-valued fitness value, where each number
    represents a different property of the molecule, for example,
    ``[1, -10, 1]``

    One way to convert the vector-valued fitness value into a
    scalar fitness value is by summing the elements, and the result in
    this case would be ``-8``. Clearly this doesn't work, because the
    resulting fitness value is not a positive number. To fix this,
    the ``-10`` should be shifted to a positive value.

    :class:`.ShiftUp` finds the minimum value of each element in the
    vector-valued fitness value across the entire population, and for
    elements where this minimum value is less than ``0``, shifts up
    the element value for every molecule in the population, so that the
    minimum value in the entire population is ``1``.

    For example, take a population with the vector-valued fitness
    values

    .. code-block:: python

            [1, -5, 5]
            [3, -10, 2]
            [2, 20, 1]

    After normalization the fitness values will be.

    .. code-block:: python

            [1, 6, 5]
            [3, 1, 2]
            [2, 31, 1]

    :class:`.ShiftUp` also works when the fitness value is a
    single value.

    Examples
    --------
    *Ensuring Positive Fitness Values*

    Here you final fitness value is calculated by taking a
    :class:`.Sum` of the different components of the fitness value.
    To ensure that the final sum is positive, each component must
    also be positive.

    .. testcode:: ensuring-positive-fitness-values

        import stk
        import numpy as np

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(1, -2, 3),
                normalized=False,
            ),
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(4, 5, -6),
                normalized=False,
            ),
        )

        # Create the normalizer.
        shifter = stk.ShiftUp()

        normalized_population = tuple(shifter.normalize(population))
        normalized_record1, normalized_record2 = normalized_population
        assert np.all(np.equal(
            normalized_record1.get_fitness_value(),
            (1, 1, 10),
        ))
        assert np.all(np.equal(
            normalized_record2.get_fitness_value(),
            (4, 8, 1),
        ))

    *Selectively Normalizing Fitness Values*

    Sometimes, you only want to normalize some members of a population,
    for example if some do not have an assigned fitness value,
    because the fitness calculation failed for whatever reason.
    You can use the `filter` parameter to exclude records from the
    normalization

    .. testcode:: selectively-normalizing-fitness-values

        import stk
        import numpy as np

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(1, -2, 3),
                normalized=False,
            ),
            # This will have a fitness value of None.
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ),
        )

        normalizer = stk.ShiftUp(
            # Only normalize values which are not None.
            filter=lambda population, record:
                record.get_fitness_value() is not None,
        )
        normalized_population = tuple(normalizer.normalize(population))
        normalized_record1, normalized_record2 = normalized_population
        assert np.all(np.equal(
            normalized_record1.get_fitness_value(),
            (1, 1, 3),
        ))
        assert normalized_record2.get_fitness_value() is None

    """

    def __init__(self, filter=lambda population, record: True):
        """
        Initialize a :class:`.ShiftUp` instance.

        Parameters
        ----------
        filter : :class:`callable`, optional
            Takes two parameters, first is a :class:`tuple`
            of :class:`.MoleculeRecord` instances,
            and the second is a :class:`.MoleculeRecord`. The
            :class:`callable` returns ``True`` or ``False``. Only
            molecules which return ``True`` will have fitness values
            normalized. By default, all molecules will have fitness
            values normalized.
            The instance passed to the `population` argument of
            :meth:`.normalize` is passed as the first argument, while
            the second argument will be passed every
            :class:`.MoleculeRecord` in it, one at a time.

        """

        self._filter = filter

    def normalize(self, population):
        filtered = filter(
            partial(self._filter, population),
            population,
        )
        # Get all the fitness arrays in a matrix.
        fmat = np.array([record.get_fitness_value() for record in filtered])

        # Get the minimum value of each element across the population.
        # keepdims ensures that np.min returns a 1-D array, because
        # it will be True if fitness values are scalar and False if
        # they are array-valued.
        mins = np.min(fmat, axis=0, keepdims=len(fmat.shape) == 1)

        # Convert all elements in mins which are not to be shifted to 0
        # and make the shift equal to the minimum value + 1.
        shift = np.zeros(len(mins))
        for i, min_ in enumerate(mins):
            if min_ <= 0:
                shift[i] = 1 - min_

        for record in population:
            if self._filter(population, record):
                yield record.with_fitness_value(
                    fitness_value=record.get_fitness_value() + shift,
                )
            else:
                yield record



"""
Replace Fitness
===============

"""

from .fitness_normalizer import FitnessNormalizer


class ReplaceFitness(FitnessNormalizer):
    """
    Replaces fitness values of a certain value with a new value.

    Examples
    --------
    *Giving a Fitness Value to Failed Calculations*

    You want to replace fitness values which are ``None`` with
    half the minimum fitness value in the population. A fitness value
    may be ``None`` because the fitness calculation failed for some
    reason.

    .. testcode:: giving-a-fitness-value-to-failed-calculations

        import stk

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=2,
                normalized=False,
            ),
            # This will have a fitness value of None.
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ),
        )

        def get_minimum_fitness_value(population):
            return min(
                record.get_fitness_value() for record in population
                if record.get_fitness_value() is not None
            )

        replacer = stk.ReplaceFitness(
            # The replacement is half the minimum fitness value in the
            # population.
            get_replacement=lambda population:
                get_minimum_fitness_value(population) / 2,
            # Only replace fitness values which are None.
            filter=lambda population, record:
                record.get_fitness_value() is None,
        )

        # Use the replacer.
        normalized_population = tuple(replacer.normalize(population))
        normalized_record1, normalized_record2 = normalized_population
        assert normalized_record1.get_fitness_value() == 2
        assert normalized_record2.get_fitness_value() == 1

    """

    def __init__(
        self,
        get_replacement,
        filter=lambda population, record: True,
    ):
        """
        Initialize a :class:`.ReplaceFitness` instance.

        Parameters
        ----------
        get_replacement : :class:`callable`
            Takes a single parameter, a :class:`tuple` of
            :class:`.MoleculeRecord` instances, which
            needs to be normalized. The entire population passed to
            :meth:`.normalize` is passed to this parameter, regardless
            of what is passed to the `filter` parameter. The
            :class:`callable` returns the value which is to be given
            to the normalized records.

        filter : :class:`callable`, optional
            Takes two parameters, first is a :class:`tuple`
            of :class:`.MoleculeRecord` instances,
            and the second is a :class:`.MoleculeRecord`. The
            :class:`callable` returns ``True`` or ``False``. Only
            molecules which return ``True`` will have fitness values
            normalized. By default, all molecules will have fitness
            values normalized.
            The instance passed to the `population` argument of
            :meth:`.normalize` is passed as the first argument, while
            the second argument will be passed every
            :class:`.MoleculeRecord` in it, one at a time.

        """

        self._get_replacement = get_replacement
        self._filter = filter

    def normalize(self, population):
        replacement = self._get_replacement(population)
        for record in population:
            if self._filter(population, record):
                yield record.with_fitness_value(replacement)
            else:
                yield record



"""
Bond Info
=========

"""


class BondInfo:
    """
    Holds additional info about :class:`.ConstructedMolecule` bonds.

    """

    def __init__(self, bond, building_block, building_block_id):
        """
        Initialize an :class:`.BondInfo` instance.

        Parameters
        ----------
        bond : :class:`.Bond`
            The bond about which information is held.

        building_block : :class:`.Molecule` or :class:`NoneType`
            The building block from which the bond originates.
            Can be ``None``, if the bond was not part of a building
            block, but was added by the construction process instead.

        building_block_id : :class:`int` or :class:`NoneType`
            A unique id for each :class:`.Molecule` placed during
            the construction of the :class:`.ConstructedMolecule`. As a
            single :class:`.Molecule` can be placed multiple times
            during construction, the `building_block_id` allows
            the user to distinguish between each placement. Can be
            ``None``, if the bond was not part of a building block, but
            was added by the construction process instead.

        """
        self._bond = bond
        self._building_block = building_block
        self._building_block_id = building_block_id

    def get_bond(self):
        """
        Get the bond about which information is held.

        Returns
        -------
        :class:`.Bond`
            The bond.

        """

        return self._bond

    def get_building_block(self):
        """
        Get the building block from which the bond originates.

        Returns
        -------
        :class:`.Molecule`
            The building block.

        None : :class:`NoneType`
            If the bond was not originally found in a building block,
            but was added by the construction process instead.

        """

        return self._building_block

    def get_building_block_id(self):
        """
        Get the id of the bond's building block.

        A unique id for each :class:`.Molecule` placed during
        the construction of the :class:`.ConstructedMolecule`. As a
        single :class:`.Molecule` can be placed multiple times
        during construction, the building block id  allows
        the user to distinguish between each placement.

        Returns
        -------
        :class:`int`
            The id.

        None : :class:`NoneType`
            If the bond was not originally found in a building block,
            but was added by the construction process instead.

        """

        return self._building_block_id



"""
M8L6 Cube
=========

"""

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import NonLinearVertex


class M8L6Cube(Cage):
    """
    Represents a cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        iron_atom = stk.BuildingBlock(
            smiles='[Fe+2]',
            functional_groups=(
                stk.SingleAtom(stk.Fe(0, charge=2))
                for i in range(6)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles='C1=NC(C=NBr)=CC=C1',
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#35]',
                    bonders=(1, ),
                    deleters=(),
                ),
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        complex = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.OctahedralDelta(
                metals=iron_atom,
                ligands=bb2,
                optimizer=stk.MCHammer(),
            ),
        )

        # Assign Bromo functional groups to the metal complex.
        iron_oct_delta = stk.BuildingBlock.init_from_molecule(
            molecule=complex,
            functional_groups=[stk.BromoFactory()],
        )

        # Define building blocks.
        bb4 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC(=C(C=C2C3=CC=C(C=C3)Br)C4=CC=C(C=C'
                '4)Br)C5=CC=C(C=C5)Br)Br'
            ),
            functional_groups=[stk.BromoFactory()],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M8L6Cube(
                building_blocks=(iron_oct_delta, bb4),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.Collapser` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        iron_atom = stk.BuildingBlock(
            smiles='[Fe+2]',
            functional_groups=(
                stk.SingleAtom(stk.Fe(0, charge=2))
                for i in range(6)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles='C1=NC(C=NBr)=CC=C1',
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#35]',
                    bonders=(1, ),
                    deleters=(),
                ),
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        complex = stk.ConstructedMolecule(
            topology_graph=stk.metal_complex.OctahedralDelta(
                metals=iron_atom,
                ligands=bb2,
                optimizer=stk.MCHammer(),
            ),
        )

        # Assign Bromo functional groups to the metal complex.
        iron_oct_delta = stk.BuildingBlock.init_from_molecule(
            molecule=complex,
            functional_groups=[stk.BromoFactory()],
        )

        # Define building blocks.
        bb4 = stk.BuildingBlock(
            smiles=(
                'C1=CC(=CC=C1C2=CC(=C(C=C2C3=CC=C(C=C3)Br)C4=CC=C(C=C'
                '4)Br)C5=CC=C(C=C5)Br)Br'
            ),
            functional_groups=[stk.BromoFactory()],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M8L6Cube(
                building_blocks=(iron_oct_delta, bb4),
                optimizer=stk.Collapser(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    Metal building blocks with three functional groups are
    required for this topology.

    Ligand building blocks with four functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 3-functional groups: 0 to 7
        | 4-functional groups: 8 to 13

    See :class:`.Cage` for more details and examples.

    """

    _non_linears = (
        NonLinearVertex(
            id=0,
            position=[1, 1, 1],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=1,
            position=[1, -1, 1],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=2,
            position=[-1, -1, 1],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=3,
            position=[-1, 1, 1],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=4,
            position=[1, 1, -1],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=5,
            position=[1, -1, -1],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=6,
            position=[-1, -1, -1],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=7,
            position=[-1, 1, -1],
            use_neighbor_placement=False,
        ),
    )

    _vertex_prototypes = (
        *_non_linears,
        NonLinearVertex(
            id=8,
            position=[0, 0, 1],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=9,
            position=[1, 0, 0],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=10,
            position=[0, 1, 0],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=11,
            position=[-1, 0, 0],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=12,
            position=[0, 0, -1],
            use_neighbor_placement=False,
        ),
        NonLinearVertex(
            id=13,
            position=[0, -1, 0],
            use_neighbor_placement=False,
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[8]),
        Edge(1, _vertex_prototypes[1], _vertex_prototypes[8]),
        Edge(2, _vertex_prototypes[2], _vertex_prototypes[8]),
        Edge(3, _vertex_prototypes[3], _vertex_prototypes[8]),
        Edge(4, _vertex_prototypes[4], _vertex_prototypes[9]),
        Edge(5, _vertex_prototypes[5], _vertex_prototypes[9]),
        Edge(6, _vertex_prototypes[1], _vertex_prototypes[9]),
        Edge(7, _vertex_prototypes[0], _vertex_prototypes[9]),
        Edge(8, _vertex_prototypes[4], _vertex_prototypes[10]),
        Edge(9, _vertex_prototypes[0], _vertex_prototypes[10]),
        Edge(10, _vertex_prototypes[3], _vertex_prototypes[10]),
        Edge(11, _vertex_prototypes[7], _vertex_prototypes[10]),
        Edge(12, _vertex_prototypes[3], _vertex_prototypes[11]),
        Edge(13, _vertex_prototypes[2], _vertex_prototypes[11]),
        Edge(14, _vertex_prototypes[6], _vertex_prototypes[11]),
        Edge(15, _vertex_prototypes[7], _vertex_prototypes[11]),
        Edge(16, _vertex_prototypes[5], _vertex_prototypes[12]),
        Edge(17, _vertex_prototypes[4], _vertex_prototypes[12]),
        Edge(18, _vertex_prototypes[7], _vertex_prototypes[12]),
        Edge(19, _vertex_prototypes[6], _vertex_prototypes[12]),
        Edge(20, _vertex_prototypes[1], _vertex_prototypes[13]),
        Edge(21, _vertex_prototypes[5], _vertex_prototypes[13]),
        Edge(22, _vertex_prototypes[6], _vertex_prototypes[13]),
        Edge(23, _vertex_prototypes[2], _vertex_prototypes[13]),
    )

    _num_windows = 4
    _num_window_types = 1



"""
Generation
==========

"""


class Generation:
    """
    An abstract base class for EA generations.

    Notes
    -----
    You might notice that the public methods of this abstract base
    class are implemented. This is just a default implementation, which
    you can ignore or override when implementing subclasses.

    """

    def __init__(
        self,
        molecule_records,
        mutation_records,
        crossover_records,
    ):
        """
        Initialize a :class:`.Generation` instance.

        Parameters
        ----------
        molecule_records : :class:`tuple` of \
                :class:`.MoleculeRecord`
            The records of molecules in the generation.

        mutation_records : :class:`tuple` of \
                :class:`.MutationRecord`
            The records of mutations done during the generation.

        crossover_records : :class:`tuple` of \
                :class:`.CrossoverRecord`
            The records of crossover operations done during the
            generation.

        """

        self._molecule_records = molecule_records
        self._mutation_records = mutation_records
        self._crossover_records = crossover_records

    def get_molecule_records(self):
        """
        Yield the molecule records in the generation.

        Yields
        ------
        :class:`.MoleculeRecord`
            A molecule record.

        """

        yield from self._molecule_records

    def get_mutation_records(self):
        """
        Yield the mutation records in the generation.

        Yields
        ------
        :class:`.MutationRecord`
            A mutation record.

        """

        yield from self._mutation_records

    def get_crossover_records(self):
        """
        Yield the crossover records in the generation.

        Yields
        ------
        :class:`.CrossoverRecord`
            A crossover record.

        """

        yield from self._crossover_records



"""
Evolutionary Algorithm
======================

"""

import logging

from stk._internal.key_makers.inchi import Inchi

from ..fitness_normalizers.null import NullFitnessNormalizer
from .implementations.parallel import Parallel
from .implementations.serial import Serial

logger = logging.getLogger(__name__)


class EvolutionaryAlgorithm:
    """
    An abstract base class for evolutionary algorithms.

    Notes
    -----
    You might notice that the public methods of this abstract base
    class are implemented. This is purely for convenience, so that
    there is a default evolutionary algorithm implementation that
    users can use. However, feel free to override the default
    implementation when implementing subclasses.

    If you do want to use the default implementation, here is a
    summary of the roles of the different components:

    .. image:: https://i.imgur.com/hGXboaU.png

    Examples
    --------
    *Subclass Implementation*

    The source code of this class can work as a good example. There
    is only one method that a subclass of
    :class:`.EvolutionaryAlgorithm` needs to implement,
    :meth:`.get_generations`, which yields :class:`.Generation`
    instances. These correspond to the generations of your
    evolutionary algorithm implementation.

    *Usage*

    There are a couple of tutorials on how to use the
    :class:`.EvolutionaryAlgorithm`, which can be found in the sidebar.

    """

    def __init__(
        self,
        initial_population,
        fitness_calculator,
        mutator,
        crosser,
        generation_selector,
        mutation_selector,
        crossover_selector,
        fitness_normalizer=NullFitnessNormalizer(),
        key_maker=Inchi(),
        num_processes=None,
    ):
        """
        Initialize a :class:`EvolutionaryAlgorithm` instance.

        Parameters
        ----------
        initial_population : :class:`tuple` of :class:`.MoleculeRecord`
            The initial population the EA should use.

        fitness_calculator : :class:`.FitnessCalculator`
            Calculates fitness values.

        mutator : :class:`.MoleculeMutator`
            Carries out mutation operations.

        crosser : :class:`.MoleculeCrosser`
            Carries out crossover operations.

        generation_selector : :class:`.Selector`
            Selects the next generation.

        mutation_selector : :class:`.Selector`
            Selects molecules for mutation.

        crossover_selector : :class:`.Selector`
            Selects molecules for crossover.

        fitness_normalizer : :class:`.FitnessNormalizer`
            Normalizes fitness values.

        key_maker : :class:`.MoleculeKeyMaker`, optional
            Used to detect duplicate molecules in the EA. If two
            molecules in a generation return the same key, one of them
            is removed.

        num_processes : :class:`int`, optional
            The number of parallel processes the EA should create.
            If ``None``, all available cores will be used.

        """

        if num_processes == 1:
            self._implementation = Serial(
                initial_population=initial_population,
                fitness_calculator=fitness_calculator,
                mutator=mutator,
                crosser=crosser,
                generation_selector=generation_selector,
                mutation_selector=mutation_selector,
                crossover_selector=crossover_selector,
                fitness_normalizer=fitness_normalizer,
                key_maker=key_maker,
                logger=logger,
            )

        else:
            self._implementation = Parallel(
                initial_population=initial_population,
                fitness_calculator=fitness_calculator,
                mutator=mutator,
                crosser=crosser,
                generation_selector=generation_selector,
                mutation_selector=mutation_selector,
                crossover_selector=crossover_selector,
                fitness_normalizer=fitness_normalizer,
                key_maker=key_maker,
                logger=logger,
                num_processes=num_processes,
            )

    def get_generations(self, num_generations):
        """
        Yield the generations of the evolutionary algorithm.

        Parameters
        ----------
        num_generations : :class:`int`
            The number of generations which should be yielded.
            Note that the initial population counts as a generation.

        Yields
        ------
        :class:`.Generation`
            A generation.

        """

        yield from self._implementation.get_generations(
            num_generations=num_generations,
        )



class FitnessNormalizer:
    """
    Abstract base class for fitness normalizers.

    A fitness normalizer takes an :class:`tuple` of
    :class:`.MoleculeRecord` instances and yields new
    :class:`.MoleculeRecord` instances, with normalized fitness values.
    The primary benefit of a normalizer vs a
    :class:`.FitnessCalculator` is that a :class:`.FitnessNormalizer`
    has access to all members in the population when it is calculating
    the normalized fitness value, whereas a :class:`.FitnessCalculator`
    does not.

    """

    def normalize(self, population):
        """
        Normalize the fitness values in `population`.

        Parameters
        ----------
        population : :class:`tuple` of :class:`.MoleculeRecord`
            The molecules which need to have their fitness values
            normalized.

        Yields
        ------
        :class:`.MoleculeRecord`
            A record with a normalized fitness value.

        """

        raise NotImplementedError()



"""
Evolutionary Algorithm Implementation
=====================================

"""

import itertools as it

from stk._internal.utilities.utilities import dedupe

from ...generation import Generation


class Implementation:
    """
    An implementation of the default evolutionary algorithm.

    """

    def __init__(
        self,
        initial_population,
        fitness_calculator,
        mutator,
        crosser,
        generation_selector,
        mutation_selector,
        crossover_selector,
        fitness_normalizer,
        key_maker,
        logger,
    ):
        """
        Initialize an :class:`.Implementation` instance.

        """

        self._initial_population = initial_population
        self._fitness_calculator = fitness_calculator
        self._mutator = mutator
        self._crosser = crosser
        self._generation_selector = generation_selector
        self._mutation_selector = mutation_selector
        self._crossover_selector = crossover_selector
        self._fitness_normalizer = fitness_normalizer
        self._key_maker = key_maker
        self._logger = logger

    def _get_generations(self, num_generations, map_):
        def get_mutation_record(batch):
            return self._mutator.mutate(batch[0])

        def get_key(record):
            return self._key_maker.get_key(record.get_molecule())

        population = self._initial_population

        self._logger.info("Calculating fitness values of initial population.")
        population = tuple(self._with_fitness_values(map_, population))
        population = tuple(
            self._fitness_normalizer.normalize(
                population=population,
            )
        )
        yield Generation(
            molecule_records=population,
            mutation_records=(),
            crossover_records=(),
        )

        for generation in range(1, num_generations):
            self._logger.info(f"Starting generation {generation}.")
            self._logger.info(f"Population size is {len(population)}.")

            self._logger.info("Doing crossovers.")
            crossover_records = tuple(self._get_crossover_records(population))

            self._logger.info("Doing mutations.")
            mutation_records = tuple(
                record
                for record in map(
                    get_mutation_record,
                    self._mutation_selector.select(population),
                )
                if record is not None
            )

            self._logger.info("Calculating fitness values.")

            offspring = (
                record.get_molecule_record() for record in crossover_records
            )
            mutants = (
                record.get_molecule_record() for record in mutation_records
            )

            population = tuple(
                self._with_fitness_values(
                    map_=map_,
                    population=tuple(
                        dedupe(
                            iterable=it.chain(population, offspring, mutants),
                            key=get_key,
                        )
                    ),
                )
            )
            population = tuple(self._fitness_normalizer.normalize(population))

            population = tuple(
                molecule_record
                for molecule_record, in self._generation_selector.select(
                    population
                )
            )

            yield Generation(
                molecule_records=population,
                mutation_records=mutation_records,
                crossover_records=crossover_records,
            )

    def _get_crossover_records(self, population):
        for batch in self._crossover_selector.select(population):
            yield from self._crosser.cross(batch)

    def _with_fitness_values(self, map_, population):
        fitness_values = map_(
            self._fitness_calculator.get_fitness_value,
            population,
        )
        for record, fitness_value in zip(population, fitness_values):
            yield record.with_fitness_value(
                fitness_value=fitness_value,
                normalized=False,
            )



"""
Power
=====

"""

import numpy as np

from .fitness_normalizer import FitnessNormalizer


class Power(FitnessNormalizer):
    """
    Raises fitness values to some power.

    Examples
    --------
    *Raising Fitness Values to a Power*

    Sometimes you might calculate a property for a molecule, where
    that property indicates a low fitness value. You can use
    :class:`.Power` to raise it to the power of -1 to get your
    final fitness value

    .. testcode:: raising-fitness-values-to-a-power

        import stk

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=1,
                normalized=False,
            ),
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=2,
                normalized=False,
            ),
        )

        normalizer = stk.Power(-1)
        normalized_population = tuple(normalizer.normalize(population))
        normalized_record1, normalized_record2 = normalized_population
        assert normalized_record1.get_fitness_value() == 1
        assert normalized_record2.get_fitness_value() == 0.5

    *Raising Fitness Values by a Set of Powers*

    In this example, assume that each fitness value consists of a
    :class:`tuple` of numbers, each representing a different property
    of the molecule, and each contributing to the final fitness value.
    The properties can be anything, such as  energy, number of atoms
    or diameter.

    If your final fitness value depends on the combination of these
    properties, you will probably first want to scale them with
    :class:`.DivideByMean`. Once this is done, you may want to
    raise each property by some power. For example
    if you raise the value of one property by ``1`` and another
    by ``-1``, it means that when the value of property 1 is big,
    the fitness value should also be big, but if the value of property
    2 is big, the fitness value should be small.

    Giving a concrete example

    .. testcode:: raising-fitness-values-by-a-set-of-powers

        import stk
        import numpy as np

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(2, 2, 2),
                normalized=False,
            ),
        )

        normalizer = stk.Power((1, -1, 2))
        normalized_population = tuple(normalizer.normalize(population))
        normalized_record, = normalized_population
        assert np.all(np.equal(
            normalized_record.get_fitness_value(),
            (2, 0.5, 4),
        ))

    *Selectively Normalizing Fitness Values*

    Sometimes, you only want to normalize some members of a population,
    for example if some do not have an assigned fitness value,
    because the fitness calculation failed for whatever reason.
    You can use the `filter` parameter to exclude records from the
    normalization

    .. testcode:: selectively-normalizing-fitness-values

        import stk
        import numpy as np

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(2, 2, 2),
                normalized=False,
            ),
            # This will have a fitness value of None.
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ),
        )

        normalizer = stk.Power(
            power=(1, -1, 2),
            # Only normalize values which are not None.
            filter=lambda population, record:
                record.get_fitness_value() is not None,
        )
        # Calling normalizer.normalize() will return a new
        # population holding the molecule records with normalized
        # fitness values.
        normalized_population = tuple(normalizer.normalize(
            population=population,
        ))
        normalized_record1, normalized_record2 = normalized_population
        assert np.all(np.equal(
            normalized_record1.get_fitness_value(),
            (2, 0.5, 4),
        ))
        assert normalized_record2.get_fitness_value() is None

    """

    def __init__(
        self,
        power,
        filter=lambda population, record: True,
    ):
        """
        Initialize a :class:`.Power` instance.

        Parameters
        ----------
        power : :class:`float` or \
                :class:`tuple` of :class:`float`
            The power each fitness value is raised to. Can
            be a single number or multiple numbers, depending on the
            form of the fitness value.

        filter : :class:`callable`, optional
            Takes two parameters, first is a :class:`tuple`
            of :class:`.MoleculeRecord` instances,
            and the second is a :class:`.MoleculeRecord`. The
            :class:`callable` returns ``True`` or ``False``. Only
            molecules which return ``True`` will have fitness values
            normalized. By default, all molecules will have fitness
            values normalized.
            The instance passed to the `population` argument of
            :meth:`.normalize` is passed as the first argument, while
            the second argument will be passed every
            :class:`.MoleculeRecord` in it, one at a time.

        """

        self._power = power
        self._filter = filter

    def normalize(self, population):
        for record in population:
            if self._filter(population, record):
                yield record.with_fitness_value(
                    fitness_value=np.float_power(
                        record.get_fitness_value(),
                        self._power,
                    ),
                )
            else:
                yield record



"""
Remove Molecules
================

"""

from .selector import Selector


class RemoveMolecules(Selector):
    """
    Prevents a :class:`.Selector` from selecting some molecules.

    Examples
    --------
    *Removing Molecules From Selection*

    .. testcode:: removing-molecules-from-selection

        import stk

        selector = stk.RemoveMolecules(
            # Do not select any molecules from the top 5 batches
            # of size 3.
            remover=stk.Best(num_batches=5, batch_size=3),
            # Select the ret of the molecules using roulette.
            selector=stk.Roulette(num_batches=20, batch_size=3),
        )
        population = tuple(
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(i)
            for i in range(10)
        )
        for batch in selector.select(population):
            # Do stuff with batch. The batch is guaranteed not to
            # contain any molecules which are found in the best 5
            # batches of size 3.
            pass

    """

    def __init__(self, remover, selector):
        """
        Initialize a :class:`.RemoveMolecules` instance.

        Parameters
        ----------
        remover : :class:`.Selector`
            Selects batches molecules, any molecule selected cannot
            be selected by `selector`.

        selector : :class:`.Selector`
            Selects batches of molecules, not containing any molecules
            selected by `remover`.

        """

        self._remover = remover
        self._selector = selector

    def select(
        self,
        population,
        included_batches=None,
        excluded_batches=None,
    ):
        remover_batches = self._remover.select(
            population=population,
            included_batches=included_batches,
            excluded_batches=excluded_batches,
        )
        removed = {record for batch in remover_batches for record in batch}
        valid_population = tuple(
            record for record in population if record not in removed
        )
        yield from self._selector.select(
            population=valid_population,
            included_batches=included_batches,
            excluded_batches=excluded_batches,
        )



"""
Remove Batches
==============

"""

from .selector import Selector


class RemoveBatches(Selector):
    """
    Prevents a :class:`.Selector` from selecting some batches.

    Examples
    --------
    *Removing Batches From Selection*

    You want to use :class:`.Roulette` selection on all but the
    5 :class:`.Worst` batches

    .. testcode:: removing-batches-from-selection

        import stk

        selector = stk.RemoveBatches(
            remover=stk.Worst(5),
            selector=stk.Roulette(20),
        )
        population = tuple(
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(i)
            for i in range(100)
        )
        for batch in selector.select(population):
            # Do stuff with batch. It was selected with roulette
            # selection and is not one of the worst 5 batches.
            pass

    """

    def __init__(self, remover, selector):
        """
        Initialize a :class:`.RemoveBatches` instance.

        Parameters
        ----------
        remover : :class:`.Selector`
            Selects batches of molecules, which cannot be yielded by
            `selector`.

        selector : :class:`.Selector`
            Selects batches of molecules, except those selected by
            `remover`.

        """

        self._remover = remover
        self._selector = selector

    def select(
        self,
        population,
        included_batches=None,
        excluded_batches=None,
    ):
        removed_batches = {
            batch.get_identity_key()
            for batch in self._remover.select(
                population=population,
                included_batches=included_batches,
                excluded_batches=excluded_batches,
            )
        }
        if excluded_batches is not None:
            removed_batches |= excluded_batches

        yield from self._selector.select(
            population=population,
            included_batches=included_batches,
            excluded_batches=removed_batches,
        )



"""
Parallel Topology Graph
=======================

"""

from __future__ import annotations

import pathos

from stk._internal.construction_state.construction_state import (
    ConstructionState,
)

from .utilities import _Placement


class _Parallel:
    """
    Holds parallel implementations of topology graph methods.

    """

    def __init__(self, stages, num_processes):
        """
        Initialize a :class:`._Parallel` instance.

        Parameters
        ----------
        stages : :class:`tuple`
            A :class:`tuple` of the form ``((v1, v2, v3), (v4, v5))``,
            where each nested :class:`tuple` holds the
            :class:`.Vertex` objects used for placement in a particular
            stage.

        num_processes : :class:`int`
            The number of parallel processes to spawn.

        """

        self._stages = stages
        self._num_processes = num_processes

    def _place_building_blocks(
        self,
        state: ConstructionState,
    ) -> ConstructionState:
        with pathos.pools.ProcessPool(self._num_processes) as pool:
            for stage in self._stages:
                vertices = tuple(state.get_vertices(stage))
                building_blocks = tuple(map(state.get_building_block, stage))
                edges = tuple(map(state.get_edges, stage))
                placements = map(
                    _Placement,
                    vertices,
                    edges,
                    building_blocks,
                )
                placement_results = pool.map(
                    lambda placement: placement.get_result(),
                    placements,
                )
                state = state.with_placement_results(
                    vertices=vertices,
                    edges=edges,
                    building_blocks=building_blocks,
                    results=placement_results,
                )
        return state

    def get_num_stages(self) -> int:
        """
        Get the number of placement stages.

        Returns:

            The number of placement stages.

        """

        return len(self._stages)



"""
Filter Molecules
================

"""

from .selector import Selector


class FilterMolecules(Selector):
    """
    Allows a :class:`.Selector` to select only some molecules.

    Examples
    --------
    *Using a Selection Algorithm on a Subset of Batches*

    .. testcode:: using-a-selection-algorithm-on-a-subset-of-batches

        import stk

        selector = stk.FilterMolecules(
            # Use only molecules in the top 5 batches of size 3.
            filter=stk.Best(num_batches=5, batch_size=3),
            # Select from those molecules using roulette selection.
            selector=stk.Roulette(num_batches=20, batch_size=3),
        )

        population = tuple(
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(i)
            for i in range(10)
        )

        for batch in selector.select(population):
            # Do stuff with batch. All the molecules in the batch
            # belong to the top 5 batches of size 3. The batch
            # was selected using roulette selection.
            pass

    """

    def __init__(self, filter, selector):
        """
        Initialize a :class:`.FilterMolecules` instance.

        Parameters
        ----------
        filter : :class:`.Selector`
            Selects molecules which can be yielded by `selector`.

        selector : :class:`.Selector`
            Selects batches of molecules. The batches can only
            contain molecules yielded by `filter`.

        """

        self._filter = filter
        self._selector = selector

    def select(
        self,
        population,
        included_batches=None,
        excluded_batches=None,
    ):
        valid_batches = self._filter.select(
            population=population,
            included_batches=included_batches,
            excluded_batches=excluded_batches,
        )
        valid_population = tuple(
            record for batch in valid_batches for record in batch
        )
        yield from self._selector.select(
            population=valid_population,
            included_batches=included_batches,
            excluded_batches=excluded_batches,
        )



from collections.abc import Callable, Iterable

import numpy as np

from stk._internal.building_block import BuildingBlock
from stk._internal.ea.molecule_records.molecule import MoleculeRecord
from stk._internal.ea.mutation.record import MutationRecord


class RandomBuildingBlock:
    """
    Substitutes random building blocks.

    This mutator takes a :class:`.ConstructedMolecule` and substitutes
    the building blocks with one chosen at random from a given set.

    Examples:

        *Constructed Molecule Mutation*

        .. testcode:: constructed-molecule-mutation

            import stk

            # Create a molecule which is to be mutated.
            bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
            bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
            polymer = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear((bb1, bb2), 'AB', 3),
            )

            # Create molecules used to substitute building blocks.
            building_blocks = (
                stk.BuildingBlock(
                    smiles='NC[Si]CCN',
                    functional_groups=[stk.PrimaryAminoFactory()],
                ),
                stk.BuildingBlock(
                    smiles='NCCCCCCCN',
                    functional_groups=[stk.PrimaryAminoFactory()],
                ),
                stk.BuildingBlock(
                    smiles='NC1CCCCC1N',
                    functional_groups=[stk.PrimaryAminoFactory()],
                ),
            )

            # Create the mutator.

            def has_primary_amino_group(building_block):
                fg, = building_block.get_functional_groups(0)
                return type(fg) is stk.PrimaryAmino

            random_bb = stk.RandomBuildingBlock(
                building_blocks=building_blocks,
                is_replaceable=has_primary_amino_group,
            )

            # Mutate a molecule.
            mutation_record1 = random_bb.mutate(polymer)

            # Mutate the molecule a second time.
            mutation_record2 = random_bb.mutate(polymer)

    """

    def __init__(
        self,
        building_blocks: Iterable[BuildingBlock],
        is_replaceable: Callable[[BuildingBlock], bool],
        name: str = "RandomBuildingBlock",
        random_seed: int | np.random.Generator | None = None,
    ) -> None:
        """
        Parameters:
            building_blocks (list[BuildingBlock]):
                A group of molecules which are used to replace building
                blocks in molecules being mutated.

            is_replaceable:
                This function is applied to every building block in
                the molecule being mutated. Building blocks
                which returned ``True`` are liable for substitution
                by one of the molecules in `building_blocks`.

            name:
                A name to help identify the mutator instance.

            random_seed:
                The random seed to use.
        """
        if random_seed is None or isinstance(random_seed, int):
            random_seed = np.random.default_rng(random_seed)

        self._building_blocks = tuple(building_blocks)
        self._is_replaceable = is_replaceable
        self._name = name
        self._generator = random_seed

    def mutate(self, record: MoleculeRecord) -> MutationRecord[MoleculeRecord]:
        """
        Return a mutant of `record`.

        Parameters:
            record:
                The molecule to be mutated.

        Returns:
            A record of the mutation.
        """
        # Choose the building block which undergoes mutation.
        replaceable_building_blocks = tuple(
            filter(
                self._is_replaceable,
                (
                    bb
                    for bb in record.get_molecule().get_building_blocks()
                    # TODO: this is actually a type error -- maybe
                    # get_building_blocks needs to return BuildingBlock
                    # instances?
                    if isinstance(bb, BuildingBlock)
                ),
            )
        )
        replaced_building_block = self._generator.choice(
            a=replaceable_building_blocks,  # type: ignore
        )
        # Choose a replacement building block.
        replacement = self._generator.choice(
            self._building_blocks,  # type: ignore
        )

        # Build the new ConstructedMolecule.
        graph = record.get_topology_graph().with_building_blocks(
            {
                replaced_building_block: replacement,
            }
        )
        return MutationRecord(
            molecule_record=MoleculeRecord(graph),
            mutator_name=self._name,
        )



"""
Normalizer Sequence
===================

"""

from .fitness_normalizer import FitnessNormalizer


class NormalizerSequence(FitnessNormalizer):
    """
    Applies other normalizers in sequence.

    Examples
    --------
    *Using Multiple Fitness Normalizers*

    You want to apply multiple fitness normalizations in sequence,
    for example, by first using :class:`.DivideByMean`, followed by
    :class:`.Sum`

    .. testcode:: using-multiple-fitness-normalizers

        import stk
        import numpy as np

        building_block = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )

        population = (
            stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(building_block, ),
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            ).with_fitness_value(
                fitness_value=(1, 100, 1000),
                normalized=False,
            ),
        )

        # Create the normalizer.
        normalizer = stk.NormalizerSequence(
            fitness_normalizers=(
                stk.DivideByMean(),
                stk.Sum(),
            ),
        )
        normalized_population = normalizer.normalize(population)
        normalized_record, = normalized_population
        assert normalized_record.get_fitness_value() == 3

    """

    def __init__(self, fitness_normalizers):
        """
        Initialize a :class:`.NormalizerSequence`.

        Parameters
        ----------
        fitness_normalizers : :class:`iterable`
            The :class:`.FitnessNormalizer` instances which should be
            used in sequence.

        """

        self._fitness_normalizers = tuple(fitness_normalizers)

    def normalize(self, population):
        for normalizer in self._fitness_normalizers:
            population = tuple(normalizer.normalize(population))
        yield from population
